#include <stdio.h>
#include <string.h>
#include "kana_spelling.h"
#include "config.h"
#include "console.h"

typedef enum {
    v_l = 1,
    v_a, v_i, v_u, v_e, v_o,
    v_k,
    v_s,
    v_t,
    v_n,
    v_h,
    v_m,
    v_y,
    v_r,
    v_w,

    v_g,
    v_z,
    v_d,
    v_b,
    v_p,

    v_c,
    v_f,
    v_j,
    v_q,
    v_v,
    v_minus,
} alpha_value_t;

#define SHIFT 5
#define ALPHA_VALUE_MAX ((v_l<<(SHIFT*3))+(v_t<<(SHIFT*2))+(v_s<<SHIFT)+v_u)
typedef struct {
    unsigned char c;
    alpha_value_t av;
} c2av_t;
c2av_t c2av_table[] = {
    {'a', v_a},
    {'i', v_i},
    {'u', v_u},
    {'e', v_e},
    {'o', v_o},
    {'k', v_k},
    {'s', v_s},
    {'t', v_t},
    {'n', v_n},
    {'h', v_h},
    {'m', v_m},
    {'y', v_y},
    {'r', v_r},
    {'w', v_w},

    {'g', v_g},
    {'z', v_z},
    {'d', v_d},
    {'b', v_b},
    {'p', v_p},

    {'c', v_c},
    {'f', v_f},
    {'j', v_j},
    {'l', v_l},
    {'q', v_q},
    {'v', v_v},
    {'-', v_minus},
};
typedef struct {
    char *hiragana;
    char *katakana;
    // ʤɡҤ餬ʤȥʤĹۤʤ礬
    int hira_len;
    int kata_len;
} kana_t;
typedef struct {
    alpha_value_t av;
    kana_t kana;
} av2kana_t;
av2kana_t av2kana_table[] = {
    {v_a,      {"", "", 0}},
    {v_i,      {"", "", 0}},
    {v_u,      {"", "", 0}},
    {v_e,      {"", "", 0}},
    {v_o,      {"", "", 0}},
    {(v_k<<SHIFT)+v_a,  {"", "", 0}},
    {(v_k<<SHIFT)+v_i,  {"", "", 0}},
    {(v_k<<SHIFT)+v_u,  {"", "", 0}},
    {(v_k<<SHIFT)+v_e,  {"", "", 0}},
    {(v_k<<SHIFT)+v_o,  {"", "", 0}},
    {(v_s<<SHIFT)+v_a,  {"", "", 0}},
    {(v_s<<SHIFT)+v_i,  {"", "", 0}},
    {(v_s<<SHIFT)+v_u,  {"", "", 0}},
    {(v_s<<SHIFT)+v_e,  {"", "", 0}},
    {(v_s<<SHIFT)+v_o,  {"", "", 0}},
    {(v_t<<SHIFT)+v_a,  {"", "", 0}},
    {(v_t<<SHIFT)+v_i,  {"", "", 0}},
    {(v_t<<SHIFT)+v_u,  {"", "", 0}},
    {(v_t<<SHIFT)+v_e,  {"", "", 0}},
    {(v_t<<SHIFT)+v_o,  {"", "", 0}},
    {(v_n<<SHIFT)+v_a,  {"", "", 0}},
    {(v_n<<SHIFT)+v_i,  {"", "", 0}},
    {(v_n<<SHIFT)+v_u,  {"", "", 0}},
    {(v_n<<SHIFT)+v_e,  {"", "", 0}},
    {(v_n<<SHIFT)+v_o,  {"", "", 0}},
    {(v_h<<SHIFT)+v_a,  {"", "", 0}},
    {(v_h<<SHIFT)+v_i,  {"", "", 0}},
    {(v_h<<SHIFT)+v_u,  {"", "", 0}},
    {(v_h<<SHIFT)+v_e,  {"", "", 0}},
    {(v_h<<SHIFT)+v_o,  {"", "", 0}},
    {(v_m<<SHIFT)+v_a,  {"", "", 0}},
    {(v_m<<SHIFT)+v_i,  {"", "", 0}},
    {(v_m<<SHIFT)+v_u,  {"", "", 0}},
    {(v_m<<SHIFT)+v_e,  {"", "", 0}},
    {(v_m<<SHIFT)+v_o,  {"", "", 0}},
    {(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_y<<SHIFT)+v_o,  {"", "", 0}},
    {(v_r<<SHIFT)+v_a,  {"", "", 0}},
    {(v_r<<SHIFT)+v_i,  {"", "", 0}},
    {(v_r<<SHIFT)+v_u,  {"", "", 0}},
    {(v_r<<SHIFT)+v_e,  {"", "", 0}},
    {(v_r<<SHIFT)+v_o,  {"", "", 0}},
    {(v_w<<SHIFT)+v_a,  {"", "", 0}},
    {(v_w<<SHIFT)+v_o,  {"", "", 0}},

    {(v_g<<SHIFT)+v_a,  {"", "", 0}},
    {(v_g<<SHIFT)+v_i,  {"", "", 0}},
    {(v_g<<SHIFT)+v_u,  {"", "", 0}},
    {(v_g<<SHIFT)+v_e,  {"", "", 0}},
    {(v_g<<SHIFT)+v_o,  {"", "", 0}},
    {(v_z<<SHIFT)+v_a,  {"", "", 0}},
    {(v_z<<SHIFT)+v_i,  {"", "", 0}},
    {(v_z<<SHIFT)+v_u,  {"", "", 0}},
    {(v_z<<SHIFT)+v_e,  {"", "", 0}},
    {(v_z<<SHIFT)+v_o,  {"", "", 0}},
    {(v_d<<SHIFT)+v_a,  {"", "", 0}},
    {(v_d<<SHIFT)+v_i,  {"", "", 0}},
    {(v_d<<SHIFT)+v_u,  {"", "", 0}},
    {(v_d<<SHIFT)+v_e,  {"", "", 0}},
    {(v_d<<SHIFT)+v_o,  {"", "", 0}},
    {(v_b<<SHIFT)+v_a,  {"", "", 0}},
    {(v_b<<SHIFT)+v_i,  {"", "", 0}},
    {(v_b<<SHIFT)+v_u,  {"", "", 0}},
    {(v_b<<SHIFT)+v_e,  {"", "", 0}},
    {(v_b<<SHIFT)+v_o,  {"", "", 0}},
    {(v_p<<SHIFT)+v_a,  {"", "", 0}},
    {(v_p<<SHIFT)+v_i,  {"", "", 0}},
    {(v_p<<SHIFT)+v_u,  {"", "", 0}},
    {(v_p<<SHIFT)+v_e,  {"", "", 0}},
    {(v_p<<SHIFT)+v_o,  {"", "", 0}},

    {(v_n<<SHIFT)+v_n,  {"", "", 0}},
    {v_minus,      {"", "", 0}},

    {(v_f<<SHIFT)+v_a,  {"դ", "ե", 0}},
    {(v_f<<SHIFT)+v_i,  {"դ", "ե", 0}},
    {(v_f<<SHIFT)+v_u,  {"", "", 0}},
    {(v_f<<SHIFT)+v_e,  {"դ", "ե", 0}},
    {(v_f<<SHIFT)+v_o,  {"դ", "ե", 0}},
    {(v_j<<SHIFT)+v_a,  {"", "", 0}},
    {(v_j<<SHIFT)+v_i,  {"", "", 0}},
    {(v_j<<SHIFT)+v_u,  {"", "", 0}},
    {(v_j<<SHIFT)+v_e,  {"", "", 0}},
    {(v_j<<SHIFT)+v_o,  {"", "", 0}},
    {(v_l<<SHIFT)+v_a,  {"", "", 0}},
    {(v_l<<SHIFT)+v_i,  {"", "", 0}},
    {(v_l<<SHIFT)+v_u,  {"", "", 0}},
    {(v_l<<SHIFT)+v_e,  {"", "", 0}},
    {(v_l<<SHIFT)+v_o,  {"", "", 0}},
    {(v_v<<SHIFT)+v_a,  {"", "", 0}},
    {(v_v<<SHIFT)+v_i,  {"", "", 0}},
    {(v_v<<SHIFT)+v_u,  {"", "", 0}},
    {(v_v<<SHIFT)+v_e,  {"", "", 0}},
    {(v_v<<SHIFT)+v_o,  {"", "", 0}},
    {(v_q<<SHIFT)+v_a,  {"", "", 0}},
    {(v_q<<SHIFT)+v_i,  {"", "", 0}},
    {(v_q<<SHIFT)+v_u,  {"", "", 0}},
    {(v_q<<SHIFT)+v_e,  {"", "", 0}},
    {(v_q<<SHIFT)+v_o,  {"", "", 0}},

    {(v_b<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"Ӥ", "ӥ", 0}},
    {(v_b<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"Ӥ", "ӥ", 0}},
    {(v_b<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"Ӥ", "ӥ", 0}},
    {(v_b<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"Ӥ", "ӥ", 0}},
    {(v_b<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"Ӥ", "ӥ", 0}},
    {(v_c<<(SHIFT*2))+(v_h<<SHIFT)+v_a,  {"", "", 0}},
    {(v_c<<(SHIFT*2))+(v_h<<SHIFT)+v_i,  {"", "", 0}},
    {(v_c<<(SHIFT*2))+(v_h<<SHIFT)+v_u,  {"", "", 0}},
    {(v_c<<(SHIFT*2))+(v_h<<SHIFT)+v_e,  {"", "", 0}},
    {(v_c<<(SHIFT*2))+(v_h<<SHIFT)+v_o,  {"", "", 0}},
    {(v_h<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"Ҥ", "ҥ", 0}},
    {(v_h<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"Ҥ", "ҥ", 0}},
    {(v_h<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"Ҥ", "ҥ", 0}},
    {(v_h<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"Ҥ", "ҥ", 0}},
    {(v_h<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"Ҥ", "ҥ", 0}},
    {(v_s<<(SHIFT*2))+(v_h<<SHIFT)+v_a,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_h<<SHIFT)+v_i,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_h<<SHIFT)+v_u,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_h<<SHIFT)+v_e,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_h<<SHIFT)+v_o,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"", "", 0}},
    {(v_k<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_k<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"", "", 0}},
    {(v_k<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_k<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"", "", 0}},
    {(v_k<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"", "", 0}},
    {(v_g<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_g<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"", "", 0}},
    {(v_g<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_g<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"", "", 0}},
    {(v_g<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"", "", 0}},

    {(v_m<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"ߤ", "ߥ", 0}},
    {(v_m<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"ߤ", "ߥ", 0}},
    {(v_m<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"ߤ", "ߥ", 0}},
    {(v_m<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"ߤ", "ߥ", 0}},
    {(v_m<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"ߤ", "ߥ", 0}},
    {(v_n<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"ˤ", "˥", 0}},
    {(v_n<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"ˤ", "˥", 0}},
    {(v_n<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"ˤ", "˥", 0}},
    {(v_n<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"ˤ", "˥", 0}},
    {(v_n<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"ˤ", "˥", 0}},

    {(v_r<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_r<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"ꤣ", "ꥣ", 0}},
    {(v_r<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_r<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"ꤧ", "ꥧ", 0}},
    {(v_r<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"", "", 0}},




    {(v_q<<(SHIFT*2))+(v_w<<SHIFT)+v_a,  {"", "", 0}},
    {(v_q<<(SHIFT*2))+(v_w<<SHIFT)+v_i,  {"", "", 0}},
    {(v_q<<(SHIFT*2))+(v_w<<SHIFT)+v_u,  {"", "", 0}},
    {(v_q<<(SHIFT*2))+(v_w<<SHIFT)+v_e,  {"", "", 0}},
    {(v_q<<(SHIFT*2))+(v_w<<SHIFT)+v_o,  {"", "", 0}},

    {(v_t<<(SHIFT*2))+(v_s<<SHIFT)+v_u,  {"", "", 0}},
    {(v_l<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_l<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_l<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"", "", 0}},
    {(v_l<<(SHIFT*2))+(v_t<<SHIFT)+v_u,  {"", "", 0}},
    {(v_l<<(SHIFT*3))+(v_t<<(SHIFT*2))+(v_s<<SHIFT)+v_u,  {"", "", 0}},
};

#define IsVowel(v) ((v)<=v_o)
#define DEFAULT_CURSOR '_'

alpha_value_t alpha2value[256];
kana_t *value2kana[ALPHA_VALUE_MAX+1];
kana_t *kana_nn;
kana_t *kana_ltu;

char *kana_buf;
int buf_maxlen = 0;
char *kana_tail;
char *buf_tail;
alpha_value_t current_av_sum;
int is_katakana = 0;
int cursor_chr = DEFAULT_CURSOR;

char *kana2str(kana_t *kana, int *len);

/*
void cat_nonroman(char key);
void cat_key(char key);
void cat_kana(kana_t *kana);
char *find_last_letter(void);
*/


void kana_spelling_init(void){
	int i;

    for (i = 0 ; i < sizeof(c2av_table)/sizeof(c2av_t) ; i++){
        assert(alpha2value[c2av_table[i].c] == 0);
        //        printf("c2v %c %d\n", c2av_table[i].c, c2av_table[i].av);
        alpha2value[c2av_table[i].c] = c2av_table[i].av;
    }
    for (i = 0 ; i < sizeof(av2kana_table)/sizeof(av2kana_t) ; i++){
        //        printf("v2k %d %s\n", av2kana_table[i].av, av2kana_table[i].kana.hiragana);
        av2kana_table[i].kana.hira_len = strlen(av2kana_table[i].kana.hiragana);
        av2kana_table[i].kana.kata_len = strlen(av2kana_table[i].kana.katakana);
        value2kana[av2kana_table[i].av] = &av2kana_table[i].kana; 

       if (av2kana_table[i].av == (v_n<<SHIFT)+v_n){
           kana_nn = &av2kana_table[i].kana;
       }
       else if (av2kana_table[i].av == (v_l<<(SHIFT*2))+(v_t<<SHIFT)+v_u){
           kana_ltu = &av2kana_table[i].kana;
        }
    }
}

char *kana_spelling_conv(char *roman, int *roman_len, int *kana_len){
	int i;
    alpha_value_t new_av, av_sum = 0;
	kana_t *kana;
	char *kana_str;

	printf("kana_spelling_conv |%s|%d\n", roman, *roman_len);
	for (i = 0 ; i < *roman_len ; i++){
	    new_av = alpha2value[(unsigned char)*roman];
		// Ƭ'n'ǼʸҲʤ''˳ꤹ
	    if (av_sum == v_n && new_av != v_n && new_av != v_y &&
		    !IsVowel(new_av))
	    {
	 	    (*roman_len) --;
	 	    kana_str = kana2str(kana_nn, kana_len);
			return kana_str;
	    }
		// ƱҲ³''
		if (av_sum == new_av && !IsVowel(new_av) && new_av != v_n){
	 	    (*roman_len) --;
			kana_str = kana2str(kana_ltu, kana_len);
			return kana_str;
	    }

		av_sum = (av_sum << SHIFT) + new_av;
		if (av_sum > ALPHA_VALUE_MAX) return NULL;
	    kana = value2kana[av_sum];
		if (kana != NULL){
			*roman_len -= i+1;
			kana_str = kana2str(kana, kana_len);
			return kana_str;
		}
		roman ++;
	}
	return NULL;
}
char *kana2str(kana_t *kana, int *len){
	if (is_katakana){
		*len = kana->kata_len;
		return kana->katakana;
	}
	else {
		*len = kana->hira_len;
		return kana->hiragana;
	}
}

int kana_spelling_is_katakana(void){
    return is_katakana;
}
void kana_spelling_set_katakana(void){
    is_katakana = -1;
}
void kana_spelling_set_hiragana(void){
    is_katakana = 0;
}

/*
void kana_spelling_init(char *buf, int maxlen, int chr){
    int i;
    buf_maxlen = maxlen;
    cursor_chr = chr;
    kana_buf = kana_tail = buf_tail = buf;
    for (i = 0 ; i < sizeof(c2av_table)/sizeof(c2av_t) ; i++){
        assert(alpha2value[c2av_table[i].c] == 0);
        //        printf("c2v %c %d\n", c2av_table[i].c, c2av_table[i].av);
        alpha2value[c2av_table[i].c] = c2av_table[i].av;
    }
    for (i = 0 ; i < sizeof(av2kana_table)/sizeof(av2kana_t) ; i++){
        //        printf("v2k %d %s\n", av2kana_table[i].av, av2kana_table[i].kana.hiragana);
        av2kana_table[i].kana.hira_len = strlen(av2kana_table[i].kana.hiragana);
        av2kana_table[i].kana.kata_len = strlen(av2kana_table[i].kana.katakana);
        value2kana[av2kana_table[i].av] = &av2kana_table[i].kana; 

       if (av2kana_table[i].av == (v_n<<SHIFT)+v_n){
           kana_nn = &av2kana_table[i].kana;
       }
       else if (av2kana_table[i].av == (v_l<<(SHIFT*2))+(v_t<<SHIFT)+v_u){
           kana_ltu = &av2kana_table[i].kana;
        }
    }
}
// Ѵȯ(result_bufʸɲä줿)0֤
int kana_spelling_input_key(unsigned char c){
    alpha_value_t new_av, new_av_sum;
    kana_t *kana;

    //    printf("in:%d-%d=%d %d\n", buf_tail, kana_buf, buf_tail-kana_buf, buf_maxlen);
    if (buf_tail - kana_buf > buf_maxlen) return 0;
    if (c == ' '){
        cat_nonroman(c);
        return 1;
    }
    new_av = alpha2value[c];
    //    printf("%c->%d %d\n", c, new_av, current_av_sum);
    if (current_av_sum == v_n && new_av != v_n && new_av != v_y &&
        !IsVowel(new_av))
    {
        cat_kana(kana_nn);
    }
    if (current_av_sum == new_av && !IsVowel(new_av) && current_av_sum != v_n){
        cat_kana(kana_ltu);
    }
    new_av_sum = (current_av_sum << SHIFT) + new_av;
    if (new_av_sum > ALPHA_VALUE_MAX){
        return 0;
    }
    current_av_sum = new_av_sum;
    //    printf("%c->%d %d\n", c, new_av, current_av);
    kana = value2kana[current_av_sum];
    if (kana != NULL){
        cat_kana(kana);
    }
    else {
        cat_key(c);
    }
    return 0;
}

void kana_spelling_backspace(void){
    //    printf("backspace:%d %d\n", buf_tail, kana_tail);
    if (buf_tail == kana_buf) return;
    if (buf_tail > kana_tail){
        buf_tail --;
        *buf_tail = cursor_chr;
        *(buf_tail+1) = '\0';
        current_av_sum = current_av_sum >> SHIFT;
    }
    else {
        kana_tail = find_last_letter();
        kana_spelling_clear_key();
    }
}
void kana_spelling_erase_cursor(void){
    int len = strlen(kana_buf);
    if (kana_buf[len-1] == cursor_chr){
        // Υʸ򥫥å
        kana_buf[len-1] = '\0';
    }
}
void kana_spelling_append_input_line(char *str){
    strcat(kana_buf + strlen(kana_buf), str);
    buf_tail = kana_tail = kana_buf + strlen(kana_buf);
    *buf_tail = cursor_chr;
    *(buf_tail+1) = '\0';
}
void kana_spelling_clear_buf(void){
    kana_tail = kana_buf;
    kana_spelling_clear_key();
}
void kana_spelling_clear_key(void){
    buf_tail = kana_tail;
    *buf_tail = cursor_chr;
    *(buf_tail+1) = '\0';
    current_av_sum = 0;
}
*/


/*
void cat_nonroman(char key){
    // nǤä֤ǥڡ褿֤פ˳ꤵ
    if (current_av_sum == v_n){
        cat_kana(kana_nn);
    }
    *buf_tail = key;
    buf_tail ++;
    kana_tail = buf_tail;
    kana_spelling_clear_key();
}
void cat_key(char key){
    *buf_tail = key;
    buf_tail ++;
    *buf_tail = '\0';
    *buf_tail = cursor_chr;
    *(buf_tail+1) = '\0';
}
void cat_kana(kana_t *kana){
    if (is_katakana){
        strcpy((char *)kana_tail, kana->katakana);
        kana_tail += kana->kata_len;
    }
    else {
        strcpy((char *)kana_tail, kana->hiragana);
        kana_tail += kana->hira_len;
    }
    buf_tail = kana_tail;
    *buf_tail = cursor_chr;
    *(buf_tail+1) = '\0';
    //    printf("|%s|(%d)|%s|\n", kana->hiragana, kana->len, kana_buf);
    current_av_sum = 0;
}
char *find_last_letter(void){
    char *p = kana_buf, *prev;
    while (p < kana_tail){
        prev = p;
        if (Is2bChr1((unsigned char)*p)){
            p += 2;
        }
        else {
            p ++;
        }
    }
    return prev;
}
*/
